# coding: utf-8
import os
from os.path import join
import sys
import shutil
import logging
import subprocess
import zc.buildout
from anybox.recipe.openerp import devtools
from anybox.recipe.openerp.base import BaseRecipe

logger = logging.getLogger(__name__)

SERVER_COMMA_LIST_OPTIONS = ('log_handler', )


class ServerRecipe(BaseRecipe):
    """Recipe for server install and config
    """
    archive_filenames = {
        '5.0': 'openerp-server-%s.tar.gz',
        '6.0': 'openerp-server-%s.tar.gz',
        '6.1': 'openerp-%s.tar.gz',
    }
    archive_nightly_filenames = {
        '6.1': 'openerp-6.1-%s.tar.gz',
        '7.0': 'openerp-7.0-%s.tar.gz',
        'trunk': 'openerp-8.0dev-%s.tar.gz'
    }
    recipe_requirements = ('babel',)
    requirements = ('pychart', 'anybox.recipe.openerp')
    soft_requirements = ('openerp-command',)
    with_openerp_command = False
    with_gunicorn = False
    with_upgrade = True
    ws = None
    template_upgrade_script = os.path.join(os.path.dirname(__file__),
                                           'upgrade.py.tmpl')

    def __init__(self, *a, **kw):
        super(ServerRecipe, self).__init__(*a, **kw)
        opt = self.options
        self.with_devtools = (
            opt.get('with_devtools', 'false').lower() == 'true')
        self.with_upgrade = self.options.get('upgrade_script') != ''
        # discarding, because we have a special behaviour with custom
        # interpreters
        opt.pop('interpreter', None)
        self.openerp_scripts = {}
        self.missing_deps_instructions.update({
            'openerp-command': ("Please provide it with 'develop' or "
                                "'gp.vcsdevelop'. "
                                "You may download it on "
                                "https://launchpad.net/openerp-command."),
        })

    def apply_version_dependent_decisions(self):
        """Store some booleans depending on detected version.

        Also does some options normalization accordingly.
        """
        gunicorn = self.options.get('gunicorn', '').strip().lower()
        self.with_gunicorn = bool(gunicorn)

        if gunicorn and self.major_version == (6, 1):
            entries = dict(direct='core', proxied='proxied')
            self.gunicorn_entry = entries.get(gunicorn)
            assert self.gunicorn_entry is not None, (
                "In OpenERP 6.1, gunicorn option value must be "
                "one of %r" % entries.keys())
        elif self.major_version >= (6, 2) and gunicorn == 'proxied':
            self.options['options.proxy_mode'] = 'True'
            logger.warn("'gunicorn = proxied' now superseded in this OpenERP "
                        "version by the 'proxy_mode' OpenERP server option ")

        self.with_openerp_command = (
            (self.with_devtools and self.major_version >= (6, 2)
             or self.major_version >= (8, 0)))

    def merge_requirements(self):
        """Prepare for installation by zc.recipe.egg

         - add Pillow iff PIL not present in eggs option.
         - (OpenERP >= 6.1) develop the openerp distribution and require it
         - gunicorn's related dependencies if needed

        For PIL, extracted requirements are not taken into account. This way,
        if at some point,
        OpenERP introduce a hard dependency on PIL, we'll still install Pillow.
        The only case where PIL will have precedence over Pillow will thus be
        the case of a legacy buildout.
        See https://bugs.launchpad.net/anybox.recipe.openerp/+bug/1017252

        Once 'openerp' is required, zc.recipe.egg will take it into account
        and put it in needed scripts, interpreters etc.
        """
        if self.major_version < (6, 0):
            self.requirements.extend(
                self.archeo_requirements.get(self.major_version))

        setup_has_pil = False
        if not 'PIL' in self.options.get('eggs', '').split():
            if 'PIL' in self.requirements:
                setup_has_pil = True
                self.requirements.remove('PIL')
            self.requirements.append('Pillow')
        if self.major_version >= (6, 1):
            openerp_dir = getattr(self, 'openerp_dir', None)
            if openerp_dir is not None:  # happens in unit tests
                self.develop(openerp_dir, setup_has_pil=setup_has_pil)
            self.requirements.append('openerp')

        if self.with_gunicorn:
            self.requirements.extend(('psutil', 'gunicorn'))

        if self.with_devtools:
            self.requirements.extend(devtools.requirements)

        if self.with_openerp_command and self.major_version < (8, 0):
            self.requirements.append('openerp-command')

        BaseRecipe.merge_requirements(self)

    def _create_default_config(self):
        """Have OpenERP generate its default config file.
        """
        self.options.setdefault('options.admin_passwd', '')
        if self.major_version <= (6, 0):
            # root-path not available as command-line option
            os.chdir(join(self.openerp_dir, 'bin'))
            subprocess.check_call([self.script_path, '--stop-after-init', '-s',
                                   ])
        else:
            sys.path.extend([self.openerp_dir])
            sys.path.extend([egg.location for egg in self.ws])
            from openerp.tools.config import configmanager
            configmanager(self.config_path).save()

    def _create_gunicorn_conf(self, qualified_name):
        """Put a gunicorn_PART.conf.py script in /etc.

        Derived from the standard gunicorn.conf.py shipping with OpenERP.
        """
        gunicorn_options = dict(
            workers='4',
            timeout='240',
            max_requests='2000',
            qualified_name=qualified_name,
            bind='%s:%s' % (
                self.options.get('options.xmlrpc_interface', '0.0.0.0'),
                self.options.get('options.xmlrpc_port', '8069')
            ))

        gunicorn_prefix = 'gunicorn.'
        gunicorn_options.update((k[len(gunicorn_prefix):], v)
                                for k, v in self.options.items()
                                if k.startswith(gunicorn_prefix))

        f = open(join(self.etc, qualified_name + '.conf.py'), 'w')
        conf = """'''Gunicorn configuration script.
Generated by buildout. Do NOT edit.'''
import openerp
bind = %(bind)r
pidfile = %(qualified_name)r + '.pid'
workers = %(workers)s

if openerp.release.major_version == '6.1':
    on_starting = openerp.wsgi.core.on_starting
    try:
      when_ready = openerp.wsgi.core.when_ready
    except AttributeError: # not in current head of 6.1
      pass
    pre_request = openerp.wsgi.core.pre_request
    post_request = openerp.wsgi.core.post_request

timeout = %(timeout)s
max_requests = %(max_requests)s

openerp.multi_process = True  # needed even with only one worker
openerp.conf.server_wide_modules = ['web']
conf = openerp.tools.config
""" % gunicorn_options

        # forwarding specified options
        prefix = 'options.'
        for opt, val in self.options.items():
            if not opt.startswith(prefix):
                continue
            opt = opt[len(prefix):]
            if opt == 'log_level':
                # blindly following the sample script
                val = dict(DEBUG=10, DEBUG_RPC=8, DEBUG_RPC_ANSWER=6,
                           DEBUG_SQL=5, INFO=20, WARNING=30, ERROR=40,
                           CRITICAL=50).get(val.strip().upper(), 30)
            if opt in SERVER_COMMA_LIST_OPTIONS:
                val = [i.strip() for i in val.split(',')]

            conf += 'conf[%r] = %r' % (opt, val) + os.linesep

        preload_dbs = self.options.get('gunicorn.preload_databases',
                                       '').split()
        if preload_dbs:
            conf += os.linesep.join((
                "",
                "def post_fork(server, worker):",
                "    '''Preload databases specified in buildout conf.'''",
                "    from openerp.modules.registry import RegistryManager",
                "    preload_dbs = %r" % preload_dbs,
                "    for db_name in preload_dbs:",
                "        server.log.info('Worker loading database %r',",
                "                        db_name)",
                "        RegistryManager.get(db_name)",
                "    server.log.info('OpenERP databases %r loaded, '",
                "                    'worker ready '",
                "                    'to serve requests', preload_dbs)",
            ))

        f.write(conf)
        f.close()

    def _get_server_command(self):
        """Return a full path to the main OpenERP server command."""
        if self.major_version <= (6, 0):
            server_cmd = join('bin', 'openerp-server.py')
        else:
            server_cmd = 'openerp-server'
        return join(self.openerp_dir, server_cmd)

    def _parse_openerp_scripts(self):
        """Parse required scripts from conf."""

        scripts = self.openerp_scripts
        if not 'openerp_scripts' in self.options:
            return
        for line in self.options.get('openerp_scripts').split(os.linesep):
            line = line.split()

            naming = line[0].split('=')
            if not naming or len(naming) > 2:
                raise ValueError("Invalid script specification %r" % line[0])
            elif len(naming) == 1:
                name = '_'.join((naming[0], self.name))
            else:
                name = naming[1]
            cl_options = []
            desc = scripts[name] = dict(entry=naming[0],
                                        command_line_options=cl_options)

            opt_prefix = 'command-line-options='
            arg_prefix = 'arguments='
            for token in line[1:]:
                if token.startswith(opt_prefix):
                    cl_options.extend(token[len(opt_prefix):].split(','))
                elif token.startswith(arg_prefix):
                    desc['arguments'] = token[len(arg_prefix):]
                else:
                    raise ValueError(
                        "Invalid token for script %r: %r" % (name, token))

    def _get_or_create_script(self, entry, name=None):
        """Retrieve or create a registered script by its entry point.

        If create_name is not given, no creation will occur, will return
        None if not found.
        In all other cases, return return (script_name, desc).
        """
        for script_name, desc in self.openerp_scripts.iteritems():
            if desc['entry'] == entry:
                return script_name, desc

        if name is not None:
            desc = self.openerp_scripts[name] = dict(entry=entry)
            return name, desc

    def _register_main_startup_script(self, qualified_name):
        """Register main startup script, usually ``start_openerp`` for install.
        """
        desc = self._get_or_create_script('openerp_starter',
                                          name=qualified_name)[1]
        desc.update(
            arguments='%r, %r, version=%r' % (self._get_server_command(),
                                              self.config_path,
                                              self.major_version),
        )

        startup_delay = float(self.options.get('startup_delay', 0))

        initialization = ['']
        if self.with_devtools:
            initialization.extend((
                'from anybox.recipe.openerp import devtools',
                'devtools.load(for_tests=False)',
                ''))

        if startup_delay:
            initialization.extend(
                ('print("sleeping %s seconds...")' % startup_delay,
                 'import time',
                 'time.sleep(%f)' % startup_delay))

        desc['initialization'] = os.linesep.join((initialization))

    def _register_test_script(self, qualified_name):
        """Register the main test script for installation.
        """
        desc = self._get_or_create_script('openerp_tester',
                                          name=qualified_name)[1]
        desc.update(
            entry='openerp_starter',
            initialization=os.linesep.join((
                "from anybox.recipe.openerp import devtools",
                "devtools.load(for_tests=True)",
                "")),
            arguments='%r, %r, version=%r, just_test=True' % (
                self._get_server_command(),
                self.config_path,
                self.major_version),
        )

    def _register_upgrade_script(self, qualified_name):
        desc = self._get_or_create_script('openerp_upgrader',
                                          name=qualified_name)[1]
        script_opt = self.options.get('upgrade_script', 'upgrade.py run')
        script = script_opt.split()
        if len(script) != 2:
            # TODO add console script entry point support
            raise zc.buildout.UserError(
                ("upgrade_script option must take the form "
                 "SOURCE_FILE CALLABLE (got '%r')" % script))
        script_source_path = self.make_absolute(script[0])
        desc.update(
            entry='openerp_upgrader',
            arguments='%r, %r, %r, %r' % (
                script_source_path, script[1],
                self.config_path, self.buildout_dir),
        )

        if not os.path.exists(script_source_path):
            logger.warning("Ugrade script source %s does not exist."
                           "Initializing it for you", script_source_path)
            shutil.copy(self.template_upgrade_script, script_source_path)

    def _register_gunicorn_startup_script(self, qualified_name):
        """Register a gunicorn foreground start script for installation.

        The produced script is suitable for external process management, such
        as provided by supervisor.
        """
        desc = self._get_or_create_script('gunicorn',
                                          name=qualified_name)[1]

        gunicorn_options = {}
        gunicorn_prefix = 'gunicorn.'
        gunicorn_options.update((k[len(gunicorn_prefix):], v)
                                for k, v in self.options.items()
                                if k.startswith(gunicorn_prefix))

        gunicorn_entry_point = gunicorn_options.get('entry_point')
        if gunicorn_entry_point is None:
            if self.major_version >= (6, 2):
                # proxy vs direct now handled by an OpenERP server option
                gunicorn_entry_point = ('openerp:'
                                        'service.wsgi_server.application')
            else:
                gunicorn_entry_point = (
                    'openerp:wsgi.%s.application' % self.gunicorn_entry)

        # gunicorn's main() does not take arguments, that's why we have
        # to resort on hacking sys.argv
        desc['initialization'] = (
            "from sys import argv; argv[1:] = ['%s', '-c', '%s.conf.py']" % (
                gunicorn_entry_point, join(self.etc, qualified_name)))

    def _register_openerp_command(self, qualified_name):
        """Register https://launchpad.net/openerp-command for install.
        """
        if self.major_version < (8, 0):
            logger.warn("Installing separate openerp-command as %r. "
                        "In OpenERP 7, openerp-command used to be "
                        "an independent python distribution, ready for "
                        "development operations, but not ready for "
                        "production operation. You are supposed to make "
                        "this distribution available in some way (alternate "
                        "PyPI server, develop, gp.vcs_develop...)",
                        qualified_name)
        desc = self._get_or_create_script('oe', name=qualified_name)[1]

        # can't reuse self.addons here, because the true addons path maybe
        # different depending on addons options, such as subdir
        addons = ':'.join(self.addons_paths)
        initialization = []
        if addons is not None:
            initialization.extend((
                "import os",
                "os.environ['OPENERP_ADDONS'] = %r" % addons,
                ''))

        if self.with_devtools:
            initialization.extend((
                'from anybox.recipe.openerp import devtools',
                'devtools.load(for_tests=True)',
                ''))
        desc['initialization'] = os.linesep.join(initialization)

    def _register_cron_worker_startup_script(self, qualified_name):
        """Register the cron worker script for installation.

        This worker script has been introduced in openobject-server, rev 4184
        together with changes in the main code that it requires.
        These changes appeared in nightly build 6.1-20120530-233414.
        The worker script itself does not appear in nightly builds.
        """
        script_src = join(self.openerp_dir, 'openerp-cron-worker')
        if not os.path.isfile(script_src):
            version = self.version_detected
            if ((version.startswith('6.1-2012') and version[4:12] < '20120530')
                    or self.version_wanted == '6.1-1'):
                logger.warn(
                    "Can't use openerp-cron-worker with version %s "
                    "You have to run a separate regular OpenERP process "
                    "for cron jobs to be launched.", version)
                return

            logger.info("Cron launcher openerp-cron-worker not found in "
                        "openerp source tree (version %s). "
                        "This is expected with some nightly builds. "
                        "Using the launcher script distributed "
                        "with the recipe.", version)
            script_src = join(os.path.split(__file__)[0],
                              'openerp-cron-worker')

        desc = self._get_or_create_script('openerp_cron_worker',
                                          name=qualified_name)[1]
        desc.update(entry='openerp_cron_worker',
                    arguments='%r, %r' % (script_src, self.config_path),
                    initialization='',
                    )

    def _install_interpreter(self):
        """Install a python interpreter with a ready-made session object."""
        int_name = self.options.get('interpreter_name', None)
        if int_name == '':  # conf requires not to build an interpreter
            return
        elif int_name is None:
            int_name = 'python_' + self.name

        initialization = os.linesep.join((
            "",
            "from anybox.recipe.openerp.runtime.session import Session",
            "session = Session(%r, %r)" % (self.config_path,
                                           self.buildout_dir),
            "if len(sys.argv) <= 1:",
            "    print('To start the OpenERP working session, just do:')",
            "    print('    session.open(db=DATABASE_NAME)')",
            "    print('or, to use the database from the buildout "
            "part config:')",
            "    print('    session.open()')",
            "    print('All other options from buildout part config "
            "do apply.')",
            ""
            "    print('Then you can issue commands such as')",
            "    print(\"    "
            "    session.registry('res.users').browse(session.cr, 1, 1)\")"
            ""))

        reqs, ws = self.eggs_reqs, self.eggs_ws
        return zc.buildout.easy_install.scripts(
            reqs, ws, sys.executable, self.options['bin-directory'],
            scripts={},
            interpreter=int_name,
            initialization=initialization,
            arguments=self.options.get('arguments', ''),
            extra_paths=self.extra_paths,
            # TODO investigate these options:
            # relative_paths=self._relative_paths,
        )

    def _install_openerp_scripts(self):
        """Install scripts registered in self.openerp_scripts.

        If initialization string is not passed, one will be cooked for
          - session initialization
          - treatment of OpenERP options specific to this script, as required
            in the 'options' key of the scripts descrition (typically to
            add a database opening option to the provided script).
        """
        reqs, ws = self.eggs_reqs, self.eggs_ws

        common_init = os.linesep.join((
            "",
            "from anybox.recipe.openerp.runtime.session import Session",
            "session = Session(%r, %r)" % (self.config_path,
                                           self.buildout_dir),
        ))

        for script_name, desc in self.openerp_scripts.items():
            initialization = desc.get('initialization', common_init)
            options = desc.get('command_line_options')
            if options:
                initialization = common_init + os.linesep.join((
                    "",
                    "session.handle_command_line_options(%r)" % options))

            zc.buildout.easy_install.scripts(
                reqs, ws, sys.executable, self.bin_dir,
                scripts={desc['entry']: script_name},
                interpreter='',
                initialization=initialization,
                arguments=desc.get('arguments', ''),
                # TODO investigate these options:
                extra_paths=self.extra_paths,
                # relative_paths=self._relative_paths,
            )
            self.openerp_installed.append(join(self.bin_dir, script_name))

    def _install_startup_scripts(self):
        """install startup and control scripts.
        """
        self._parse_openerp_scripts()

        # provide additional needed entry points for main start/test scripts
        self.eggs_reqs.extend((
            ('openerp_starter',
             'anybox.recipe.openerp.runtime.start_openerp',
             'main'),
            ('openerp_cron_worker',
             'anybox.recipe.openerp.runtime.start_openerp',
             'main'),
            ('openerp_upgrader',
             'anybox.recipe.openerp.runtime.upgrade',
             'upgrade'),
        ))

        if self.major_version >= (8, 0):
            self.eggs_reqs.append(('oe', 'openerpcommand.main', 'run'))

        self._install_interpreter()

        main_script = self.options.get('script_name', 'start_' + self.name)
        self._register_main_startup_script(main_script)
        self.script_path = join(self.bin_dir, main_script)

        if self.with_openerp_command:
            self._register_openerp_command(
                self.options.get('openerp_command_name',
                                 '%s_command' % self.name))

        if self.with_devtools:
            self._register_test_script(
                self.options.get('test_script_name', 'test_' + self.name))

        if self.with_gunicorn:
            qualified_name = self.options.get('gunicorn_script_name',
                                              'gunicorn_%s' % self.name)
            self._create_gunicorn_conf(qualified_name)
            self._register_gunicorn_startup_script(qualified_name)

            qualified_name = self.options.get('cron_worker_script_name',
                                              'cron_worker_%s' % self.name)
            self._register_cron_worker_startup_script(qualified_name)

        if self.with_upgrade:
            qualified_name = self.options.get('upgrade_script_name',
                                              'upgrade_%s' % self.name)
            self._register_upgrade_script(qualified_name)

        self._install_openerp_scripts()

    def _60_fix_root_path(self):
        """Correction of root path for OpenERP 6.0 pure python install"""

        if 'options.root_path' not in self.options:
            self.options['options.root_path'] = join(self.openerp_dir, 'bin')

    archeo_requirements = {
        (5, 0): ['psycopg2', 'pytz', 'lxml', 'egenix-mx-base',
                 'reportlab', 'pydot',
                 ],
        }



import os
import sys
import subprocess
import logging
import tempfile

from ..utils import working_directory_keeper
from .base import BaseRepo
from .base import SUBPROCESS_ENV
from anybox.recipe.openerp import utils
import re

logger = logging.getLogger(__name__)


class GitRepo(BaseRepo):
    """Represent a Git clone tied to a reference branch."""

    vcs_control_dir = '.git'

    def clean(self):
        if not os.path.isdir(self.target_dir):
            return
        with working_directory_keeper:
            os.chdir(self.target_dir)
            subprocess.check_call(['git', 'clean', '-fdqx'])

    def parents(self):
        """Return full hash of parent nodes. """
        with working_directory_keeper:
            os.chdir(self.target_dir)
            p = subprocess.Popen(['git', 'rev-parse', '--verify', 'HEAD'],
                                 stdout=subprocess.PIPE, env=SUBPROCESS_ENV)
            return p.communicate()[0].split()

    def uncommitted_changes(self):
        """True if we have uncommitted changes."""
        with working_directory_keeper:
            os.chdir(self.target_dir)
            p = subprocess.Popen(['git', 'status', '--short'],
                                 stdout=subprocess.PIPE, env=SUBPROCESS_ENV)
            out = p.communicate()[0]
            return bool(out.strip())

    def get_update(self, revision):
        """Ensure that target_dir is a branch of url at specified revision.

        If target_dir already exists, does a simple fetch.
        Offline-mode: no branch nor fetch, but checkout.
        """
        target_dir = self.target_dir
        url = self.url
        offline = self.offline
        rev_str = revision

        with working_directory_keeper:
            is_target_dir_exists = os.path.exists(target_dir)
            if not is_target_dir_exists:
                # TODO case of local url ?
                if offline:
                    raise IOError(
                        "git repository %s does not exist; cannot clone it "
                        "from %s (offline mode)" % (target_dir, url))

                os.chdir(os.path.split(target_dir)[0])
                logger.info("Cloning %s ...", url)
                subprocess.check_call(['git', 'clone', url, target_dir])
            os.chdir(target_dir)

            if is_target_dir_exists:
                # TODO what if remote repo is actually local fs ?
                if not offline:
                    logger.info("Fetch for git repo %s (rev %s)...",
                                target_dir, rev_str)
                    subprocess.check_call(['git', 'fetch'])

            if revision and self._needToSwitchRevision(revision):
                # switch to the expected revision
                logger.info("Checkout %s to revision %s",
                            target_dir, revision)
                self._switch(revision)

    def archive(self, target_path):
        revision = self.parents()[0]
        if not os.path.exists(target_path):
            os.makedirs(target_path)
        with working_directory_keeper:
            os.chdir(self.target_dir)
            target_tar = tempfile.NamedTemporaryFile(
                prefix=os.path.split(self.target_dir)[1] + '.tar')
            target_tar.file.close()
            subprocess.check_call(['git', 'archive', revision,
                                   '-o', target_tar.name])
            subprocess.check_call(['tar', '-x', '-f', target_tar.name,
                                   '-C', target_path])
            os.unlink(target_tar.name)

    def _needToSwitchRevision(self, revision):
        """ Check if we need to checkout to an other branch
        """
        p = utils.check_output(['git', 'rev-parse', '--abbrev-ref', 'HEAD'])
        rev = p.split()[0]  # remove \n
        logger.info("Current revision '%s' - Expected revision '%s'",
                    rev, revision)
        return rev != revision

    def _switch(self, revision):
        rbp = self._remote_branch_prefix
        branches = utils.check_output(["git", "branch", "-a"])
        branch = revision
        if re.search("^(\*| ) %s$" % re.escape(branch), branches, re.M):
            # the branch is local, normal checkout will work
            logger.info("The branch is local; normal checkout ")
            argv = ["checkout", branch]
        elif re.search(
            "^  " + re.escape(rbp) + "\/" + re.escape(branch) + "$", branches,
                re.M):
            # the branch is not local, normal checkout won't work here
            logger.info("The branch is not local; checkout remote branch ")
            argv = ["checkout", "-b", branch, "%s/%s" % (rbp, branch)]
        else:
            # A tag or revision was specified instead of a branch
            logger.info("Checkout tag or revision")
            argv = ["checkout", revision]
        # runs the checkout with predetermined arguments
        argv.insert(0, "git")
        subprocess.check_call(argv)

    @property
    def _remote_branch_prefix(self):
        version = self._git_version
        if version < (1, 6, 3):
            return "origin"
        else:
            return "remotes/origin"

    @property
    def _git_version(self):
        out = utils.check_output(["git", "--version"])
        m = re.search("git version (\d+)\.(\d+)(\.\d+)?(\.\d+)?", out)
        if m is None:
            logger.error("Unable to parse git version output")
            logger.error("'git --version' output was:\n%s\n%s", out)
            sys.exit(1)
        version = m.groups()

        if version[3] is not None:
            version = (
                int(version[0]),
                int(version[1]),
                int(version[2][1:]),
                int(version[3][1:])
            )
        elif version[2] is not None:
            version = (
                int(version[0]),
                int(version[1]),
                int(version[2][1:])
            )
        else:
            version = (int(version[0]), int(version[1]))
        if version < (1, 5):
            logger.error(
                "Git version %s is unsupported, please upgrade",
                ".".join([str(v) for v in version]))
            sys.exit(1)
        return version



"""VCS tests: Subversion."""

import os
import subprocess
from ..testing import VcsTestCase
from ..svn import SvnCheckout


class SvnTestCase(VcsTestCase):

    def create_src(self):
        os.chdir(self.src_dir)
        subprocess.check_call(['svnadmin', 'create', 'src-repo'])
        self.src_repo = os.path.join(self.src_dir, 'src-repo')
        self.src_repo = 'file://' + self.src_repo

        tmp_checkout = os.path.join(self.src_dir, 'tmp_checkout')
        subprocess.call(['svn', 'checkout', self.src_repo, tmp_checkout])

        os.chdir(tmp_checkout)
        f = open('tracked', 'w')
        f.write("first" + os.linesep)
        f.close()
        subprocess.call(['svn', 'add', 'tracked'])
        subprocess.call(['svn', 'commit', '-m', 'initial commit'])
        f = open('tracked', 'w')
        f.write("last" + os.linesep)
        f.close()
        subprocess.call(['svn', 'commit', '-m', 'last version'])

    def test_checkout(self):
        """Svn clone."""
        target_dir = os.path.join(self.dst_dir, "Mycheckout")
        SvnCheckout(target_dir, self.src_repo)('head')

        self.assertTrue(os.path.isdir(target_dir))
        f = open(os.path.join(target_dir, 'tracked'))
        lines = f.readlines()
        f.close()
        self.assertEquals(lines[0].strip(), 'last')



import os
import shutil
import subprocess
from .. import utils

SUBPROCESS_ENV = os.environ.copy()
SUBPROCESS_ENV['PYTHONPATH'] = SUBPROCESS_ENV.pop(
    'BUILDOUT_ORIGINAL_PYTHONPATH', '')


class UpdateError(subprocess.CalledProcessError):
    """Specific class for errors occurring during updates of existing repos.
    """


class CloneError(subprocess.CalledProcessError):
    """Class to easily signal errors in initial cloning.
    """


def wrap_check_call(exc_cls, call_fn):

    def wrapped_check_call(*args, **kwargs):
        """Variant on subprocess.check_* that raises %s.""" % exc_cls
        try:
            return call_fn(*args, **kwargs)
        except subprocess.CalledProcessError, e:
            up_exc = exc_cls(e.returncode, e.cmd)
            output = getattr(e, 'output', None)
            if output is not None:
                up_exc.output = output
            raise up_exc

    return wrapped_check_call

update_check_call = wrap_check_call(UpdateError, subprocess.check_call)
clone_check_call = wrap_check_call(CloneError, subprocess.check_call)
update_check_output = wrap_check_call(UpdateError, utils.check_output)
clone_check_output = wrap_check_call(CloneError, utils.check_output)


class BaseRepo(object):

    def __init__(self, target_dir, url, clear_retry=False,
                 offline=False, clear_locks=False, **options):

        self.target_dir = target_dir
        self.url = url
        self.clear_retry = clear_retry
        self.offline = offline
        self.clear_locks = clear_locks

        # additional options that may depend on the VCS subclass
        self.options = options

    def clear_target(self):
        shutil.rmtree(self.target_dir)

    def clean(self):
        """Remove unwanted untracked files.

        The default implementation removes Python object files and
        (resulting) empty directories.
        Subclasses are supposed to implement better vcs-specific behaviours.
        It is important for release-related options that this cleaning does not
        appear as a local modification.
        """
        utils.clean_object_files(self.target_dir)

    def __call__(self, revision):
        """Create if needed from remote source, and put it at wanted revision.
        """
        if self.options.get('clean'):
            self.clean()

        try:
            self.get_update(revision)
        except UpdateError:
            if self.offline or not self.clear_retry:
                raise
            self.clear_target()
            self.get_update(revision)
        return self  # nicer in particular for tests

    @classmethod
    def is_versioned(cls, path):
        """True if path exists and is versioned under this vcs.

        Common implementation based on vcs_control_dir class attribute.
        """
        return os.path.exists(os.path.join(path, cls.vcs_control_dir))

    @classmethod
    def fix_target(cls, target_dir):
        """Take into account that some targets are actually shifted below.

        That is the case of standalon addon (see launchpad #1012899).
        """

        if os.path.exists(target_dir) and not cls.is_versioned(target_dir):
            name = os.path.split(target_dir)[-1]
            new_target = os.path.join(target_dir, name)
            manifest = os.path.join(new_target, '__openerp__.py')
            manifest2 = os.path.join(new_target, '__terp__.py')
            exists = os.path.exists(manifest)
            exists = exists or os.path.exists(manifest2)
            if cls.is_versioned(new_target) and exists:
                return new_target
        return target_dir

    def uncommitted_changes(self):
        """True if we have uncommitted changes."""
        raise NotImplementedError

    def parents(self):
        """Return universal identifier for parent nodes, aka current revisions.

        There might be more than one with some VCSes (ex: pending merge in hg).
        """
        raise NotImplementedError



"""Uniform encapsulation of buildout-local upgrade script.

The idea is to provide a common set of options, so that upgrade scripts all
have the same interface, and project maintainers can focus on the decision
taking logic.
"""
import os
import sys
import imp
import logging
from argparse import ArgumentParser
from argparse import ArgumentDefaultsHelpFormatter
from argparse import SUPPRESS
from datetime import datetime
from math import ceil

from .session import Session

DEFAULT_LOG_FILE = 'upgrade.log'


def upgrade(upgrade_script, upgrade_callable, conf, buildout_dir):
    """Run the upgrade from a source file.

    All arguments are set in the standalone script produced by buildout through
    entry point options.

    * ``upgrade_script``: absolute path to the upgrade script python source.
    * ``upgrade_callable``: name of the callable in source file actually
      running the script.

      It must accept the two following positional arguments, in that order:

        - a :class:`.Session` instance (as in standard "OpenERP scripts")
        - a logger (standard object from the :mod:`logging` module)

      and may return a non zero status code to indicate an error.
      Both ``None`` and 0 are interpreted as success.

    * ``conf``: path to the OpenERP configuration file (managed by the recipe)
    * ``buildout_dir``: directory of the buildout
    """

    parser = ArgumentParser(formatter_class=ArgumentDefaultsHelpFormatter,
                            epilog="")
    parser.add_argument('--log-file', default=DEFAULT_LOG_FILE,
                        help="File to log sub-operations to, relative to the "
                        "current working directory, supports homedir "
                        "expansion ('~' on POSIX systems).")
    parser.add_argument('--log-level', default='info',
                        help="Main OpenERP logging level. Does not affect the "
                        "logging from the main upgrade script itself.")
    parser.add_argument('--console-log-level', default='info',
                        help="Level for the upgrade process console "
                        "logging. This is for the main upgrade script itself "
                        "meaning that usually only major steps should be "
                        "logged ")
    parser.add_argument('-q', '--quiet', action='store_true',
                        help="Suppress console output from the main upgrade "
                             "script (lower level stages can still write)")
    parser.add_argument('-d', '--db-name', default=SUPPRESS,
                        help="Database name. If ommitted, the general default "
                        "values from OpenERP config file or libpq will apply.")
    parser.add_argument('--init-load-demo-data', action='store_true',
                        help="Demo data will be loaded with module "
                        "installations if and only if "
                        "this modifier is specified")

    arguments = parser.parse_args()  # 'args' would shadow the one of pdb
    log_path = os.path.abspath(os.path.expanduser(arguments.log_file))
    log_level = arguments.log_level
    console_level = arguments.console_log_level.upper()
    quiet = arguments.quiet

    try:
        log_file = open(log_path, 'a')
    except IOError:
        sys.stderr.write("Cannot open %r for write" % log_path + os.linesep)
        sys.exit(-1)

    session = Session(conf, buildout_dir)

    from openerp.tools import config
    config['logfile'] = log_path
    config['log-level'] = log_level

    start_time = datetime.utcnow()
    if not quiet:
        print("Starting upgrade, logging details to %s at level %s, "
              "and major steps to console at level %s" % (
                  log_path, log_level.upper(), console_level.upper()))
        print('')

    logger = logging.getLogger('openerp.upgrade')
    console_handler = logging.StreamHandler()
    console_handler.setLevel(getattr(logging, console_level))
    console_handler.setFormatter(logging.Formatter(
        "%(asctime)s %(levelname)s  %(message)s"))

    if not arguments.quiet:
        logger.addHandler(console_handler)

    db_name = getattr(arguments, 'db_name', None)
    logger.info("Opening database %r", db_name)
    session.open(db=db_name, with_demo=bool(arguments.init_load_demo_data))
    # actual value after all defaultings have been done
    db_name = session.cr.dbname

    if session.is_initialization:
        logger.info("Database %r base initialization done. Proceeding further",
                    db_name)
    else:
        logger.info("Database %r loaded. Actual upgrade begins.", db_name)

    pkg_version = session.package_version
    if pkg_version is None:
        logger.warn("Expected package version file %r does not exist. "
                    "version won't be set in database at the end of upgrade. "
                    "Consider including such a version file in your project "
                    "*before* version dependent logic is actually needed.",
                    session.version_file_path)
    else:
        logger.info("Read package version: %s from %s", pkg_version,
                    session.version_file_path)

    db_version = session.db_version
    if db_version is None:
        if not session.is_initialization:
            logger.warn("No version currently set in database (the present "
                        "upgrade script has never been run). Consider setting "
                        "database version even for fresh instances, to "
                        "eliminate any guesswork in the upgrade scripts.")
    else:
        logger.info("Database latest upgrade version : %s", db_version)

    upgrade_module = imp.load_source('anybox.recipe.openerp.upgrade_openerp',
                                     upgrade_script)
    statuscode = getattr(upgrade_module, upgrade_callable)(session, logger)
    if statuscode is None or statuscode == 0:
        if pkg_version is not None:
            logger.info("setting version %s in database" % pkg_version)
            session.db_version = pkg_version
        session.cr.commit()

        logger.info("%s successful. Total time: %d seconds." % (
            "Initialization" if session.is_initialization else "Upgrade",
            ceil((datetime.utcnow() - start_time).total_seconds())
        ))
    else:
        logger.error("Please check logs at %s" % log_path)

    log_file.close()
    sys.exit(statuscode)



import os
import sys
from setuptools import setup, find_packages

version = '1.8.2'

if sys.version_info < (2, 6):
    sys.stderr.write("This package requires Python 2.6 or newer. "
                     "Yours is " + sys.version + os.linesep)
    sys.exit(1)

requires = ['setuptools', 'zc.recipe.egg', 'zc.buildout']

if sys.version_info < (2, 7):
    requires.append('ordereddict')
    requires.append('argparse')

setup(
    name="anybox.recipe.openerp",
    version=version,
    author="Anybox",
    author_email="contact@anybox.fr",
    description="A buildout recipe to install and configure OpenERP",
    license="AGPLv3+",
    long_description='\n'.join((
        open('README.rst').read(),
        open('CHANGES.rst').read())),
    url="https://launchpad.net/anybox.recipe.openerp",
    packages=find_packages(),
    zip_safe=False,
    include_package_data=True,
    install_requires=requires,
    tests_require=requires + ['nose', 'bzr'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Buildout :: Recipe',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU Affero General Public License v3 or '
        'later (AGPLv3+)',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    entry_points={'zc.buildout': [
        'server = anybox.recipe.openerp.server:ServerRecipe',
        'webclient = anybox.recipe.openerp.webclient:WebClientRecipe',
        'gtkclient = anybox.recipe.openerp.gtkclient:GtkClientRecipe',
    ]},
    extras_require = {
      'bzr' : ['bzr'] },
)



from .base import UpdateError  # noqa
from .hg import HgRepo
from .bzr import BzrBranch
from .svn import SvnCheckout
from .git import GitRepo

SUPPORTED = {}

SUPPORTED['hg'] = HgRepo
SUPPORTED['bzr'] = BzrBranch
SUPPORTED['git'] = GitRepo
SUPPORTED['svn'] = SvnCheckout


def get_update(vcs_type, target_dir, url, revision, **options):
    """General entry point."""
    cls = SUPPORTED.get(vcs_type)
    if cls is None:
        raise ValueError("Unsupported VCS type: %r" % vcs_type)

    # case of standalon addon (see launchpad #1012899)
    target_dir = cls.fix_target(target_dir)
    cls(target_dir, url, **options)(revision)



import unittest
import os
import shutil
from tempfile import mkdtemp
from ..utils import working_directory_keeper

COMMIT_USER_NAME = 'Test'
COMMIT_USER_EMAIL = 'test@example.org'
COMMIT_USER_FULL = '%s %s' % (COMMIT_USER_NAME, COMMIT_USER_EMAIL)


class VcsTestCase(unittest.TestCase):
    """Common fixture"""

    def setUp(self):
        sandbox = self.sandbox = mkdtemp('test_oerp_recipe_vcs')
        src = self.src_dir = os.path.join(sandbox, 'src')
        dst = self.dst_dir = os.path.join(sandbox, 'dest')
        os.mkdir(src)
        os.mkdir(dst)
        with working_directory_keeper:
            self.create_src()

    def create_src(self):
        """Create a source repository to run most tests.

        To be implemented in subclasses."""
        raise NotImplementedError

    def tearDown(self):
        print "TEARDOWN remove " + self.sandbox
        shutil.rmtree(self.sandbox)



"""Enable multiple addons paths in OpenERP 5.0.x

This is done as monkey patch, because the recipe must take care of not
modifying the source, so that release scripts don't refuse releasing the
buildout directory.

As of this writing, version 5.0 has been frozen for long already.
"""

import sys
import os


def do_patch():

    import addons

    # reverting side effects of addons import, and storing our addons path
    if addons.ad is not None:
        addons.buildout_addons_paths = [
            p.strip() for p in addons.tools.config['addons_path'].split(',')]
        sys.path.remove(addons.ad)
        if addons._ad != addons.ad:
            sys.path.remove(addons._ad)
        addons.ad = None

    def get_module_path(module, downloaded=False):
        """Return the path of the given module.

        """

        logger, netsvc = addons.logger, addons.netsvc
        for path in addons.buildout_addons_paths:
            module_path = os.path.join(path, module)
            if (os.path.exists(module_path)
                    or os.path.exists(module_path + '.zip')):
                logger.notifyChannel('init', netsvc.LOG_INFO,
                                     'module %r: found at %s' % (module,
                                                                 module_path))

                return module_path

        if downloaded:
            return os.path.join(addons._ad, module)
        logger.notifyChannel('init', netsvc.LOG_WARNING,
                             'module %r: module not found' % (module,))
        return False

    def get_modules():
        """Returns the list of module names
        """
        def listdir(dir):
            def clean(name):
                name = os.path.basename(name)
                if name[-4:] == '.zip':
                    name = name[:-4]
                return name

            def is_really_module(name):
                name = os.path.join(dir, name)
                return os.path.isdir(name) or addons.zipfile.is_zipfile(name)

            return map(clean, filter(is_really_module, os.listdir(dir)))

        return list(set(m for path in addons.buildout_addons_paths
                        for m in listdir(path)))

    import imp
    import zipimport

    def register_class(m):
        """
        Register module named m, if not already registered

        Only modification from the original function is mentionned as comment.
        """

        logger, netsvc = addons.logger, addons.netsvc

        def log(e):
            mt = isinstance(e, zipimport.ZipImportError) and 'zip ' or ''
            msg = "Couldn't load %smodule %s" % (mt, m)
            logger.notifyChannel('init', netsvc.LOG_CRITICAL, msg)
            logger.notifyChannel('init', netsvc.LOG_CRITICAL, e)

        if m in addons.loaded:
            return

        logger.notifyChannel('init', netsvc.LOG_INFO,
                             'module %r: registering objects' % m)
        mod_path = get_module_path(m)

        try:
            zip_mod_path = mod_path + '.zip'
            if not os.path.isfile(zip_mod_path):
                # line below modified
                fm = imp.find_module(m, addons.buildout_addons_paths)
                try:
                    imp.load_module(m, *fm)
                finally:
                    if fm[0]:
                        fm[0].close()
            else:
                zimp = zipimport.zipimporter(zip_mod_path)
                zimp.load_module(m)
        except Exception, e:
            log(e)
            raise
        else:
            addons.loaded.append(m)

    addons.get_module_path = get_module_path
    addons.get_modules = get_modules
    addons.register_class = register_class
    import tools.misc

    orig_file_open = tools.misc.file_open

    def file_open(name, mode="r", subdir='addons', pathinfo=False):
        """Modified version for calls after direct join of module names.

        For example, this can be seen in ``addons.load_module_graph()``
           fp = tools.file_open(opj(m, filename))

        Found it simpler (and necessary, given there were references to
        addons_path and root_path) to correct from here.

        The original one does try and replace subpaths by zip files if
        some are found. We simply call it back to avoid duplicating this
        (in the same way it's already doing after subdir related resolutions).
        """

        if name.replace(os.path.sep, '/').startswith('addons/'):
            subdir = 'addons'
            name = name[7:]

        # First try to locate in addons_path
        if subdir:  # remember that the default value is 'addons'
            subdir2 = subdir
            if subdir2.replace(os.path.sep, '/').startswith('addons/'):
                subdir2 = subdir2[7:]

            subdir2 = (subdir2 != 'addons' or None) and subdir2
            if os.path.isabs(name):
                fn = name
            elif subdir2 is None:
                # first segment is necessary the module name
                fn = addons.get_module_resource(*name.split(os.path.sep, 1))

            try:
                fn = os.path.normpath(fn)
                # subdir=None will skip this override's equivalent in
                # original code (leading to zip handling etc)
                fo = file_open(fn, mode=mode, subdir=None, pathinfo=pathinfo)
                if pathinfo:
                    return fo, fn
                return fo
            except IOError:
                pass

        rtp = os.path.normcase(os.path.abspath(tools.config['root_path']))
        if subdir:
            name = os.path.join(rtp, subdir, name)
        else:
            name = os.path.join(rtp, name)

        name = os.path.normpath(name)
        return orig_file_open(name, subdir=None, pathinfo=pathinfo)

    tools.misc.file_open = tools.file_open = file_open
